//
// History:
//
// 27/01/99 15:30 by Alexander Trush
// 30/08/99 13:30 by Alexander Trush

#include <idc.idc>

static mask(xxadr) {

	auto shift;

	shift = 2 * (Byte(xxadr+2)&15);

	     if(shift== 0) return       "0xFF";
	else if(shift== 1) return "0xC000003F";
	else if(shift== 2) return "0xF000000F";
	else if(shift== 3) return "0xFC000003"; 
	else if(shift== 4) return "0xFF000000";
	else if(shift== 5) return "0x3FC00000";   
	else if(shift== 6) return  "0xFF00000";  
	else if(shift== 7) return  "0x3FC0000"; 
	else if(shift== 8) return   "0xFF0000";
	else if(shift== 9) return   "0x3FC000";
	else if(shift==10) return    "0xFF000";
	else if(shift==11) return    "0x3FC00";
	else if(shift==12) return     "0xFF00";
	else if(shift==13) return     "0x3FC0";
	else if(shift==14) return      "0xFF0";
	else 		   return      "0x3FC";

}

static main(void) {

  auto addr, mask;
  auto Offset;


  StringStp(0xA);
  Tabs(1);
  Comments(1);
  XrefShow(5);
  AutoShow(1);
  Indent(16);
  CmtIndent(40);

  MakeDword(0X14);
  MakeCode(0);
  Wait();

  Message("Checking for inderect branch with link\n");
  for(addr=MinEA(); addr != BADADDR; addr=NextAddr(addr)) {

	if( Dword(addr)==0xE1A0E00F ) {  // MOV LR, PC  0xE1A0E00F

		Offset = addr+8;

		if(Name(Offset)=="") MakeName(Offset,"indBL_"+atoa(Offset));
		MakeCode(addr);
		add_dref(addr, Offset, dr_T);
		OpAlt(addr,1,"PC ; ["+Name(Offset)+"]");

		if( (Dword(addr+4) & 0x0C10F000)==0x0410F000 ||     // LDR PC,
		    (Dword(addr+4) & 0x0FFFFFF0)==0x01A0F000	) { // MOV PC,
			Message("*");
			MakeCode(Offset);
		}else Message("?");

		Message("Found at %08X\n", addr );
	}
  }
  Wait();

  Message("Checking for the PC relative operand\n");
  for(addr=MinEA(); addr != BADADDR; addr=NextAddr(addr))
  {
        if( (  substr(GetMnem(addr),0,3)=="ADD" || 	  // Mnemonic without condition
               substr(GetMnem(addr),0,3)=="SUB"     ) &&
            GetOpType(addr,1)==1    &&                    // Reg
            GetOpType(addr,2)==5    &&                    // Imm
            GetOpnd(addr,1)=="PC" 
          ) {

		Offset=GetOperandValue(addr,2);

		if(  GetMnem(addr+4)==GetMnem(addr)      &&
	             GetOpType(addr+4,1)==1              &&  // Reg
	             GetOpType(addr+4,2)==5              &&  // Imm
                     GetOpnd(addr+4,0)==GetOpnd(addr,0)  &&
                     GetOpnd(addr+4,1)==GetOpnd(addr,0)
		  )  {
			//    ࠡ  PC
			Offset = Offset + GetOperandValue(addr+4,2);
			if( substr(GetMnem(addr),0,3)=="SUB" )  Offset = -Offset;
	                Offset = (addr+8) + Offset;

//			if(Name(Offset)=="") MakeName(Offset,"relpc_"+atoa(Offset));
			OpAlt(addr,2,"#("+Name(Offset)+"-.+8) & " + mask(addr));
			OpAlt(addr+4,2,"#("+Name(Offset)+"-.+4) & " + mask(addr+4));
                                                                    
			add_dref(addr, Offset, dr_T);

		}
		else {
			if( substr(GetMnem(addr),0,3)=="SUB" )  Offset = -Offset;
	                Offset = (addr+8) + Offset;


//			if(Name(Offset)=="") MakeName(Offset,"relpc_"+atoa(Offset));
			OpAlt(addr,2,"#"+Name(Offset)+"-.+8");

			add_dref(addr, Offset, dr_T);
		}


		Message("Found at %08X:%s %08X\n", addr, GetMnem(addr), Offset );

	   }

  }
  Wait();


  Message("Checking for 'forgotten' instarction\n");
  for(addr=MinEA(); addr != BADADDR; addr=NextAddr(addr)) {
	if( (Dword(addr)&0xFF000000)==0xEA000000 ) {  // Brach instruction?

		Offset=Dword(addr) & 0x00FFFFFF;
		if(    Dword(addr) & 0x00800000) Offset = Offset | -0x01000000;
		Offset=addr+8+(Offset<<2);




		if(isCode(GetFlags(Offset))) { // Test for jump to code
				MakeCode(addr);
				Message("Found branch to code at %08X\n", addr );
		} else if(Offset==addr) { // Test for forever loop
					MakeCode(addr);
					Message("Found forever loop at %08X\n", addr );
					MakeRptCmt(addr, "forever loop...");
		       }
//

		if( (Dword(addr+4)&0xFF000000)==0xEA000000 &&   // Too Branch ins
                    (Dword(addr) & 0x00FFFFFF)==((Dword(addr+4)+1) & 0x00FFFFFF) ) {

				MakeCode(addr+4);
				Message("Found double branch at %08X\n", addr );
		}
	} else if( (Dword(addr) & 0xFE108000)==0xE8108000 &&
		    Dword(addr)==Dword(addr+4)               ) {

				MakeCode(addr+4);
				Message("Found double LDM Rx, {..., PC} at %08X\n", addr );
	}
  }
  Wait();
  Message("This is all!\n");
}
